"""
ising.py
written in Python3
author: C. Lockhart <chris@lockhartlab.org>
"""


class Ising:
    def __init__(self, lattice, pbc=True, j=0., h=1.):
        self.lattice = lattice

    def energy(self):
        pass






ising = Ising()
